package uk.co.hcsoftware.spambuddy;

import uk.co.hcsoftware.yago.*;
import uk.co.hcsoftware.yago.util.DefaultArgReaders;
import uk.co.hcsoftware.yago.parsers.GnuParser;

import javax.mail.internet.*;
import javax.mail.*;
import javax.activation.DataHandler;
import javax.activation.FileDataSource;
import java.io.*;
import java.util.*;
import java.util.logging.Level;
import java.util.logging.Logger;

import ns.TaskTimer;


public class Spambuddy {
    private static final String version="1-20061224";
    public static final Logger log = Logger.getLogger(Spambuddy.class.getName());
    private Option<Integer> port;

    public static void main(String[] args) {
        new Spambuddy(args);
    }

    public Spambuddy(String[] argv) {
        try {
            setupOptions();

            gnuParser.parseOrDie("spambuddy",argv);
            if(help.isSet()){
                System.out.println(gnuParser.getUsage("spambuddy"));
                System.exit(0);
            }
            else run();
        } catch (Exception e) {
            log.log(Level.WARNING, "Exception", e);
        }
    }
    private ArgReader<List<InternetAddress>> addressListReader = new ArgReader<List<InternetAddress>>() {
        public List<InternetAddress> readArg(String argValue) throws InvalidArgException {
            try {
                List<InternetAddress> l=new ArrayList<InternetAddress>();
                //arg is either a file with a list of addresses or
                //commasep list of addresses
                File file = new File(argValue);
                if(file.exists()){
                    log.info("arg is a file containing addresses");
                    log.info("reading file="+file);
                    LineNumberReader lnr=new LineNumberReader(new FileReader(file));
                    try {
                        String line;
                        while((line=lnr.readLine())!=null){
                            if (line.length()>0 && !(line.startsWith("#") || line.startsWith("//"))) {
                                l.add(new InternetAddress(line));
                            }
                        }
                    } catch (IOException e) {
                        throw new InvalidArgException("error reading file",e,argValue);
                    } finally {
                        try {
                            lnr.close();
                        } catch (IOException e) {
                            log.log(Level.WARNING,"IOe",e);
                        }
                    }

                }

                else{ String[] paths=argValue.split(",");

                    for(String p:paths){
                        InternetAddress f = new InternetAddress(p);
                        l.add(f);
                    }
                }
                return l;

            } catch (AddressException e) {
                throw new InvalidArgException("bad inetaddress",e,argValue);
            } catch (FileNotFoundException e) {
                throw new InvalidArgException("can'to find file",e,argValue);
            }
        }
    };
    private ArgReader<InternetAddress> addressReader = new ArgReader<InternetAddress>() {
        public InternetAddress readArg(String argValue) throws InvalidArgException {
            try {
                return new InternetAddress(argValue);
            } catch (AddressException e) {
                throw new InvalidArgException("bad inetaddress",e,argValue);
            }
        }
    };
    private ArgReader<List<File>> fileListReader = new ArgReader<List<File>>() {
        public List<File> readArg(String argValue) throws InvalidArgException {
            // /var/logs/log.txt:/var/logs/logs2.txt (also check existence)
            String[] paths=argValue.split(File.pathSeparator);
            List<File> l=new ArrayList<File>(paths.length);
            for(String p:paths){
                File f = new File(p);
                if(!f.exists()) throw new InvalidArgException("file doesn'to exist",argValue);
                l.add(f);
            }
            return l;
        }
    };

    //all options are optional, args are requried
    //cc (list of cc addresses) abc@efg,gef@efg
    //bcc (list of bcc addresses) abc@efg,gef@efg
    //s subject
    //u username
    //p password
    //host host
    //r replyto
    //a attachments: /var/logs/log.txt:/var/logs/logs2.txt

    //from(from address): abc@xyz,abc

    //batchsize n send in batches of n
    //batchinterval n wait n seconds between each send

    //arg1: (list of to addresses) abc@efg,gef@efg
    private enum Field{
        t("to"),c("cc"),b("bcc");
        private final String s;

        public static Field get(String s){
            s=s.toLowerCase();
            if("to".equals(s)) return t;
            else if("cc".equals(s)) return c;
            else if("bcc".equals(s)) return b;
            return null;
        }
        Field(String s){
            this.s = s;
        }

        public static String valuesPrint(String delim){
            StringBuilder sb = new StringBuilder();
            for(Field f:values()){
                if(sb.length()>0) sb.append(delim);
                sb.append(f);
            }
            return sb.toString();
        }

        public String toString() {
            return s;
        }
    }
    private ArgReader<String> stringReader;
    private ArgReader<Integer> intReader;
    private Option<Integer> batchsize ;
    private Option<Integer> batchinterval;
    private Option<Field> batchfield ;

    private Option<List<InternetAddress>> to ,cc , bcc ;
    private Option<String> s ,u ,p ,host , r;
    private Argument<String> text ;
    private Option<File> html ;
    private Option<List<File>> a;
    private Option<InternetAddress> from ;
    private Option<Date> sendAt;
    private Option tls;

    private Option help=new Option("help",false,'h',"help",null,"prints out this usage message");
    GnuParser gnuParser;

    private void setupOptions(){
        stringReader = DefaultArgReaders.createStringReader();
        intReader = DefaultArgReaders.createIntReader();

        //batch
        batchsize = new Option<Integer>("batchsize",false,intReader,"send in batches of n");
        batchsize.setArgName("n");
        batchinterval = new Option<Integer>("batchinterval",false,intReader,"wait n seconds between each send");
        batchinterval.setArgName("n");
        batchfield = new Option<Field>("batchfield",false
                ,new ArgReader<Field>() {
                    public Field readArg(String s) throws InvalidArgException {
                        if(s.length()<1) throw new InvalidArgException("must be one of: "+Field.valuesPrint("|")+" (no arg supplied)",s);
                        final Field field = Field.get(s);
                        if(field==null) throw new InvalidArgException("must be one of: "+Field.valuesPrint("|")+"(was "+s+")",s);
                        return field;
                    }
                }, "field which is split into batches");
        batchfield.setArgName(Field.valuesPrint("|"));

         List<String> dateFormats = Arrays.asList(new String[]{
            "yyyyMMddhh:mm"
        });
        sendAt=new Option<Date>("sendat",false
                ,DefaultArgReaders.createDateReader(dateFormats),"send email at specified time, format: "+dateFormats);
        sendAt.setArgName("time");


        //addressing
        to = new Option<List<InternetAddress>>("to",false,addressListReader,"to address(es)");
        final String addressListReaderDesc = "addresses | file";
        to.setArgName(addressListReaderDesc);
        cc= new Option<List<InternetAddress>>("cc",false,addressListReader,"cc address(es) abc@efg,gef@efg");
        cc.setArgName(addressListReaderDesc);
        bcc= new Option<List<InternetAddress>>("bcc",false,addressListReader,"bcc address(es) abc@efg,gef@efg");
        bcc.setArgName(addressListReaderDesc);

        //ISP options
        host = new Option<String>("host",true,stringReader,"SMTP host");
        port = new Option<Integer>("port",false, intReader, "SMTP port");
        host.setArgName("hostname");
        p = new Option<String>("p",false,stringReader,"password");
        p.setArgName("pass");
u = new Option<String>("u",false,stringReader,"username");
        u.setArgName("userid");

        //mail content
         r = new Option<String>("r",false,stringReader,"replyto");
        r.setArgName("replyto");
         from = new Option<InternetAddress>("from",true,addressReader,"from address e.g. abc@def.org");
        from.setArgName("address");


        s = new Option<String>("s",false,stringReader,"subject");
        s.setArgName("subject");
        html = new Option<File>("html",false
                ,DefaultArgReaders.createFileReader(true),"html content for email (in addition to text content from std.in)");
        html.setArgName("file");
        text = new Argument<String>(false,stringReader,"text","email text (if not present reads from std.in)");

        a = new Option<List<File>>("a",false,fileListReader,"attachments e.g. /var/logs/log.txt"+File.pathSeparator+"/var/logs/logs2.txt");
        a.setArgName("filename(s)");

        tls = new Option("tls",false, "Use TLS (encrypt communication with server) default=false");


        Option[] options = {cc,bcc,s,u,p,host,r,a,from,batchsize,batchinterval,batchfield,to,html,sendAt,tls,help,port};
        Argument[] args = {text};

        gnuParser = new GnuParser(options,args);
        gnuParser.createBundleOptionGroup(batchsize,batchfield,batchinterval);
        gnuParser.createAtLeastOneOptionGroup(to,cc,bcc);
    }
    private void setMailHeaders(MimeMessage email) throws MessagingException {
       Hashtable ht = new Hashtable();
       if (r.isSet()) {
            email.addHeader("Reply-To",r.getArgValue());
        }
        email.addHeader("X-Mailer","Spambuddy v"+version+" http://www.hcsoftware.co.uk/software/spambuddy/");
    }
    protected Session getMailSession()
    {
            Properties properties = new Properties(System.getProperties());
//            properties.setProperty(MAIL_TRANSPORT_PROTOCOL, SMTP);

//            if (!org.apache.commons.lang.StringUtils.isNotEmpty(this.hostName))
//            {
//                throw new EmailException(
//                    "Cannot find valid hostname for mail session");
//            }
//
//            properties.setProperty(MAIL_PORT, smtpPort);
//            properties.setProperty(MAIL_HOST, hostName);
//            properties.setProperty(MAIL_DEBUG, String.valueOf(this.debug));
//
//            // changed this (back) to getInstance due to security exceptions
//            // caused when testing using maven
//            this.session = Session.getInstance(properties, this.authenticator);
        properties.setProperty("mail.transport.protocol", "smtp");
        if (host.isSet())  properties.setProperty("mail.host",host.getArgValue());
        if(port.isSet()){
            log.info("setting port="+port.getArgValue());
            properties.setProperty("mail.port",String.valueOf(port.getArgValue()));
        }
        if (from.isSet()) {
//            final InternetAddress fromAddress = from.getArgValue();
            properties.setProperty("mail.smtp.from", from.getArgValue().toString());//.setFrom(fromAddress.getAddress(),fromAddress.getPersonal());
        }

        PasswordAuthentication auth = null;
        if (u.isSet()) {
            properties.setProperty("mail.smtp.auth", "true");
            if(p.isSet()){
                auth = new PasswordAuthentication(u.getArgValue(),p.getArgValue());
//                email.setAuthentication(u.getArgValue(),p.getArgValue());
            }
            else{
                log.warning("-u set but not -p, using empty password");
               auth = new PasswordAuthentication(u.getArgValue(),"");
//               email.setAuthentication(u.getArgValue(),"");
            }
        }
        if(tls.isSet()){
            properties.setProperty("mail.imap.starttls.enable","true");
            properties.setProperty("mail.smtp.starttls.enable","true");
        }
        final PasswordAuthentication auth1 = auth;
        if(auth==null) return Session.getInstance(properties);
        else return Session.getInstance(properties,new Authenticator() {
            protected PasswordAuthentication getPasswordAuthentication() {
                return auth1;
            }
        });
    }

    public void run() throws IOException, MessagingException {
        log.entering(Spambuddy.class.getName(),"run");
        Session session = getMailSession();

        MimeMessage email = createEmail(session);//new MimeMessage(session); //email = createEmail();
        setMailHeaders(email);

        if(s.isSet()) email.setSubject(s.getArgValue());


        if(sendAt.isSet()){
            Date d = sendAt.getArgValue();
            long time = d.getTime();
            long delta = time-System.currentTimeMillis();
            if(delta<0){
                System.out.println("Abort SEND! "+sendAt.getArgName()+" is in the past!");
                System.exit(1);
            }
            log.info("waiting until "+d);
            try {
                Thread.sleep(delta);
            } catch (InterruptedException e) {
                log.log(Level.WARNING, "exception", e);
                System.out.println("Abort SEND! "+sendAt.getArgName()+" wait thread interrupted!");
                System.exit(1);
            }
        }
        if (r.isSet()) {
            email.setReplyTo(new Address[]{new InternetAddress(r.getArgValue())});
        }

        if(batchsize.isSet()){
            sendBatchedEmail(email);
        }
        else{
            if(cc.isSet()) email.setRecipients(Message.RecipientType.CC,toArray(cc.getArgValue()));
            if(bcc.isSet()) email.setRecipients(Message.RecipientType.BCC,toArray(bcc.getArgValue()));//email.setRecipients(Message.RecipientType.BCC,toArray(bcc.getArgValue()));//email.setBcc(bcc.getArgValue());
            if (to.isSet()) email.setRecipients(Message.RecipientType.TO,toArray(to.getArgValue()));//email.setTo(to.getArgValue());
            log.info("sending mail");
//            email.setPopBeforeSmtp(true,"pop.googlemail.com","nathanielspurling@googlemail.com","mailgoggle");
//            email.send();
//            if (this.popBeforeSmtp)
//            {
//                Store store = session.getStore("pop3");
//                store.connect(this.popHost, this.popUsername, this.popPassword);
//            }
            email.setSentDate(new Date());
//            email.set
            Transport.send(email);
        }
    }
    private InternetAddress[] toArray(List<InternetAddress> l){
        return l.toArray(new InternetAddress[l.size()]);
    }

    private void sendBatchedEmail(MimeMessage email) throws MessagingException {
        //send in batches
        int batchSize=batchsize.getArgValue();
        log.fine("send in batches of "+batchSize);

        final List<InternetAddress> batchList;
        Field f = batchfield.getArgValue();
        if(f==Field.t){
            batchList=to.getArgValue();
            if(cc.isSet()) email.setRecipients(Message.RecipientType.CC,toArray(cc.getArgValue()));
            if(bcc.isSet()) email.setRecipients(Message.RecipientType.BCC,toArray(bcc.getArgValue()));
        }
        else if(f==Field.c){
            batchList=cc.getArgValue();
            if (to.isSet())  email.setRecipients(Message.RecipientType.TO,toArray(to.getArgValue()));
            if(bcc.isSet()) email.setRecipients(Message.RecipientType.BCC,toArray(bcc.getArgValue()));
        }
        else{
            assert f==Field.b;
            batchList=bcc.getArgValue();
            if(cc.isSet()) email.setRecipients(Message.RecipientType.CC,toArray(cc.getArgValue()));
            if (to.isSet())  email.setRecipients(Message.RecipientType.TO,toArray(to.getArgValue()));
        }
        int batchcount = batchList.size()/batchSize;
        if(batchList.size()%batchSize>0) batchcount++;
        final int wt = batchinterval.getArgValue();
        long millis=((batchcount-1)*wt*1000);

        TaskTimer tt = new TaskTimer();

        log.info("I estimate this will take "+tt.formatMillis(millis)+": should be finished around "+new Date(System.currentTimeMillis()+millis));

        while(!batchList.isEmpty()){
            int addresscount = Math.min(batchSize,batchList.size());
            List<InternetAddress> addresses =batchList.subList(0,addresscount);
            log.info("sending email to "+addresses);
            if(f==Field.t) email.setRecipients(Message.RecipientType.TO,toArray(addresses));
            else if(f==Field.c) email.setRecipients(Message.RecipientType.CC,toArray(addresses));
            else email.setRecipients(Message.RecipientType.BCC,toArray(addresses));

            Transport.send(email);
            for(int j=0;j<addresscount;j++) batchList.remove(0);

            if(!batchList.isEmpty() && batchinterval.isSet()){
                final int waitTime = batchinterval.getArgValue();
                log.info("waiting "+waitTime+"s");
                try {
                    Thread.sleep(1000*waitTime);
                } catch (InterruptedException e) {
                    log.log(Level.WARNING,"interrupted",e);
                }
            }
        }
    }

    private String getEmailText() throws IOException {
        if (!text.wasFound()) {
            int i;
            InputStreamReader sr = new InputStreamReader(System.in);
            char[] buf=new char[256];
            StringBuffer sb = new StringBuffer();
            while((i=sr.read(buf))!=-1){
                sb.append(buf,0,i);
            }
            log.info("read "+sb.length()+" chars from std.in");
            return sb.toString();
        } else {
            return text.getArgValue();
        }
    }

    private MimeMessage createEmail(Session session) throws IOException, MessagingException {
        String s=getEmailText();
        final MimeMessage email = new MimeMessage(session);
        if(html.isSet() || a.isSet()){
            Multipart mp = new MimeMultipart("alternative");
            BodyPart bp2 = new MimeBodyPart();
            bp2.setContent(s,"text/plain");
            mp.addBodyPart(bp2);
            email.setContent(mp);

            if (html.isSet()) {
                File f = html.getArgValue();
                FileInputStream fr = new FileInputStream(f);
                byte[] buf = new byte[(int) f.length()];
                fr.read(buf);
                fr.close();
                BodyPart bp = new MimeBodyPart();
//            bp.setDataHandler(new DataHandler(new FileDataSource(html.getArgValue())));
                bp.setContent(new String(buf,"UTF8"),"text/html");
                mp.addBodyPart(bp);
            }
            if(a.isSet()){
                List<File> l = a.getArgValue();
                int i=0;
                for(File f:l){
                    MimeBodyPart mbp = new MimeBodyPart();
                    final String desc = "attachment "+(++i);
                    mbp.setDisposition(Part.ATTACHMENT);
                    mbp.setFileName(f.getName());
                    mbp.setDescription(desc);
                    mbp.setDataHandler(new DataHandler(new FileDataSource(f)));
		    mp.addBodyPart(mbp);
                }
            }

        }
        else{
            email.setContent(s,"text/plain");
        }
        return email;
    }

//    private void addAttachments(MultiPartEmail email) throws EmailException, MalformedURLException {
//        List<File> l = a.getArgValue();
//        int i=0;
//        for(File f:l){
//            final String desc = "attachment "+(++i);
//            email.attach(f.toURL(),desc,desc);
//        }
//    }

}
